from io import BytesIO

import avro.schema
from avro.io import DatumReader, DatumWriter, BinaryDecoder, BinaryEncoder

# 对应 Record.avsc 文件中269行的 Record。kafka中的一条数据就是一个Record，因此需要Record的schema来反序列化
RECORD: str = "com.tencent.subscribe.avro.Record"


class RecordDeserializer:
    # parsed_schema: 从文件中读取的所有schema
    # record_schema: com.tencent.subscribe.avro.Record 的schema，用于反序列化
    __slots__ = ['parsed_schema', 'record_schema']

    # schema文件的 绝对路径 或者 相对于该脚本所在目录的相对路径
    def __init__(self, schema_file: str):
        s = avro.schema.parse(open(schema_file, "rb").read())
        # Record.avsc 中定义了多个schema，所以解析后的类型应该是 UnionSchema 子类
        assert isinstance(s, avro.schema.UnionSchema)
        self.parsed_schema = s
        # 遍历所有的schema并找到 com.tencent.subscribe.avro.Record 的schema
        for schema in self.parsed_schema.schemas:
            if isinstance(schema, avro.schema.RecordSchema) and RECORD == schema.fullname:
                self.record_schema = schema
                break

    def deserialize(self, data: bytes, des_old) -> dict:
        try:
            return self.deserialize_inner(data, self.record_schema, self.record_schema)
        except Exception as e:
            error_msg = str(e)
            print(f"warning: deserialize avro data error: {error_msg}, will try again.")
            try:
                # try again for old version
                return self.deserialize_inner(data, des_old.record_schema, self.record_schema)
            except Exception as e:
                error_msg = str(e)
                print(f"deserialize avro old version data error: {error_msg}")

    def deserialize_inner(self, data: bytes, writers_schema, readers_schema) -> dict:
        try:
            bio = BytesIO(initial_bytes=data)
            decoder = BinaryDecoder(bio)
            res = DatumReader(writers_schema=writers_schema, readers_schema=readers_schema).read(decoder)
        except Exception:
            raise
        # Avro records are represented as Python dicts.
        assert isinstance(res, dict)
        return res

    def serialize(self, data: object) -> bytes:
        bio = BytesIO()
        encoder = BinaryEncoder(bio)
        writer = DatumWriter(writers_schema=self.record_schema)
        writer.write(data, encoder)
        return bio.getvalue()

    def get_type(self, data) -> str:
        """返回与data匹配的schema name，可能存在多个可以匹配的schema。返回的name不包含namespace"""
        for s in self.parsed_schema.schemas:
            assert isinstance(s, avro.schema.Schema)
            if s.validate(data):
                # BinaryGeometry BinaryObject TextGeometry TextObject 三种类型 validate 无法区分
                if s.name in ("BinaryGeometry", "BinaryObject", "TextGeometry", "TextObject"):
                    if isinstance(data['value'], str):
                        if data['type'] == "BinaryGeometry":
                            return "BinaryGeometry"
                        if data['type'] in ("blob", "binary"):
                            return "BinaryObject"
                        if data['type'] in ("ENUM", "SET", "json"):
                            return "TextObject"
                        return "TextGeometry"
                return s.name
        raise Exception("error to find schema for: {}".format(data))
