# -*- coding:utf-8 -*-
import base64
import json
from datetime import datetime

import kafka

import const
import deserializer
import to_sql

PARTITION_SEQ_KEY = "ps"
PARTITION_SEQ2_KEY = "ps2"
CHECKPOINT_KEY = "Checkpoint"
HEARTBEAT_KEY = "Heartbeat"
# 初始化 deserializer
des = deserializer.RecordDeserializer("Record.avsc")
# Record-v6是上一个版本的schema，当前版本在此基础上增加了oldColumnsBytes/newColumnsBytes两个字段，用于处理超大消息
des_old = deserializer.RecordDeserializer("Record-v6.avsc")


## 本Demo基本执行流程为：
# 1.创建 Kafka 消费者，不停消费消息；
# 2.依次遍历每条消息，根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理；
# 3.partitionMsgConsumer 检查消息头的序列号，如果发现重复消息则丢弃，如果发现消息丢失则终止运行；
def block_consume(brokers, topic, group, user, password, trans2sql):
    # create a kafka consumer
    if user != "":
        consumer = kafka.KafkaConsumer(topic,
                                       auto_offset_reset='earliest',
                                       enable_auto_commit=False,
                                       group_id=group,
                                       sasl_plain_username=user,
                                       sasl_plain_password=password,
                                       sasl_mechanism='SCRAM-SHA-512',
                                       security_protocol='SASL_PLAINTEXT',
                                       bootstrap_servers=brokers)
    else:
        consumer = kafka.KafkaConsumer(topic,
                                       auto_offset_reset='earliest',
                                       enable_auto_commit=False,
                                       group_id=group,
                                       bootstrap_servers=brokers)
        # get partition of topic
    partitions = consumer.partitions_for_topic(topic)
    partition_consumers = {}
    for p in partitions:
        partition_consumers[p] = PartitionMsgConsumer()
    for message in consumer:
        partition_consumers[message.partition].consume(consumer, message, trans2sql)


def get_partition_seq(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ_KEY:
            return int(m[1].decode())
    return -1


def get_partition_seq2(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ2_KEY:
            return int(m[1].decode())
    return -1


def isCheckpoint(message):
    for m in message.headers:
        if m[0] == CHECKPOINT_KEY:
            return True
    return False


def isHeartbeat(message):
    for m in message.headers:
        if m[0] == HEARTBEAT_KEY:
            return True
    return False


def timestamp2str(timestamp):
    date_time = datetime.fromtimestamp(timestamp)
    return date_time.strftime("%Y-%m-%d %H:%M:%S")


#  如果传入trans2sql的值为1，将Entry里的数据解析为sql语句并打印出来。
#  您可以在此函数基础上开发满足自己需要的数据处理逻辑。
def on_entry(partition, offset, partition_seq, record, trans2sql):
    print(
        "-->[partition: {}, offset: {}, partitionSeq: {}] [{}:{}], happenedAt: {}, record: {}".format(
            partition, offset, partition_seq, record[const.FILE_NAME],
            record[const.POSITION], timestamp2str(record[const.TIMESTAMP]), record)
    )
    if trans2sql:
        to_sql.trans2sql(record, des)


# 将形如 {'array': [{'BinaryObject': {'type': 'blob', 'value': 'f0VMRgIBAQAA...'}}]}
# 的数据转为 [{'type': 'blob', 'value': b'\x7fELF\x02\x01\x01...'}] 的格式，
# 以便使得大消息的数据格式与普通消息的数据格式保持一致，方便解析。
def large_msg_avro_data_transfer(avro_data):
    if isinstance(avro_data, dict):
        print("transfer avro_data from map to list and decode bash64")
        # old_cols = [list(item.values())[0] for item in old_cols['array']]
        cols_list = []
        for avro_item in avro_data['array']:
            inner_item = list(avro_item.values())[0]
            # 这几种类型原始数据为字节数组，序列化过程中自动做了base64转换，需要还原
            if "Character" in avro_item or "BinaryObject" in avro_item or "BinaryGeometry" in avro_item:
                try:
                    inner_item["value"] = base64.b64decode(inner_item["value"])
                except Exception as e:
                    print("warning: No base64 decoded data in Character/BinaryObject/BinaryGeometry."
                          " Exception:{}".format(e))
            cols_list.append(inner_item)
        return cols_list
    return avro_data


class PartitionMsgConsumer(object):
    def __init__(self):
        self.last_partition_seq = 0
        self.last_partition_seq2 = 0
        self.large_record = None

    def consume(self, consumer, message, trans2sql):
        # 每个消息的消息头部都有一个编号记做ps，这个编号在同一个分区内正常情况下保证单调递增且连续
        # ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
        # ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
        # 重复解析部分binlog导致的数据重复投递问题。
        ps2 = get_partition_seq2(message)
        if ps2 != -1:
            if self.last_partition_seq2 > 0:
                if ps2 < self.last_partition_seq2:
                    print("duplicated msg:{}".format(message))
                    return
                elif ps2 == self.last_partition_seq2:  # 可能是重复消息也可能是checkpoint或heartbeat消息
                    if not isCheckpoint(message) and not isHeartbeat(message):
                        print("duplicated msg:{}".format(message))
                        return
                elif ps2 != self.last_partition_seq2 + 1:
                    # 在极端情况下，kafka也可能出现丢失消息的情况（kafka自身机制的原因，概率极小），此时partitionSeq不连续，
                    # 如果对数据一致性要求较高，需要停止消费数据，然后在控制台重置订阅任务再重新配置。
                    # 如果对数据一致性要求不高，可以接受个别消息丢失，那么跳过此条件判断即可。
                    msg = "the partition seq2 discontinuous, last: {}, current: {}".format(self.last_partition_seq2,
                                                                                           ps2)
                    raise Exception(msg)
            self.last_partition_seq2 = ps2
            valid_ps = ps2
        else:  # 老的订阅任务没有ps2，会走到这里
            ps = get_partition_seq(message)
            if ps == -1:
                pass
            elif ps <= self.last_partition_seq:
                # 一些特殊情况下消息可能出现重复写入的情况，如网络异常导致的重试等，此时丢弃重复的消息即可
                print("duplicated msg:{}".format(message))
                return
            elif ps != self.last_partition_seq + 1 and self.last_partition_seq != 0:
                msg = "the partition seq discontinuous, last: {}, current: {}".format(self.last_partition_seq, ps)
                raise Exception(msg)
            self.last_partition_seq = ps
            valid_ps = ps

        # 将kafka消息的Value值转换为 Record
        record = des.deserialize(message.value, des_old)

        if record[const.VERSION] == 2 and record[const.TOTAL] > 1:
            print(
                f"merging large record, partition:{message.partition}, offset:{message.offset}, partitionSeq:{valid_ps}, "
                f"Total:{record[const.TOTAL]}, Index:{record[const.INDEX]}, GTID:{record[const.GTID]}")
            if record[const.INDEX] == 0:
                self.large_record = record
            else:
                if record[const.OLD_COLUMNS_B]:
                    if self.large_record[const.OLD_COLUMNS_B]:
                        self.large_record[const.OLD_COLUMNS_B] = self.large_record[const.OLD_COLUMNS_B] + record[
                            const.OLD_COLUMNS_B]
                    else:
                        self.large_record[const.OLD_COLUMNS_B] = record[const.OLD_COLUMNS_B]
                if record[const.NEW_COLUMNS_B]:
                    if self.large_record[const.NEW_COLUMNS_B]:
                        self.large_record[const.NEW_COLUMNS_B] = self.large_record[const.NEW_COLUMNS_B] + record[
                            const.NEW_COLUMNS_B]
                    else:
                        self.large_record[const.NEW_COLUMNS_B] = record[const.NEW_COLUMNS_B]
            if record[const.INDEX] < record[const.TOTAL] - 1:
                return
            if self.large_record[const.OLD_COLUMNS_B]:
                try:
                    print("large_record.OLD_COLUMNS_B.length: {}".format(len(self.large_record[const.OLD_COLUMNS_B])))
                    old_cols = json.loads(self.large_record[const.OLD_COLUMNS_B])
                    # 组合后的大消息字段反序列化之后的格式和正常消息有一些差异，需要转一下格式
                    formated_data = large_msg_avro_data_transfer(old_cols)
                    self.large_record[const.OLD_COLUMNS] = formated_data
                    self.large_record[const.OLD_COLUMNS_B] = b''
                except json.JSONDecodeError as err:
                    return err
            if self.large_record[const.NEW_COLUMNS_B]:
                try:
                    print("large_record.NEW_COLUMNS_B.length: {}".format(len(self.large_record[const.NEW_COLUMNS_B])))
                    new_cols = json.loads(self.large_record[const.NEW_COLUMNS_B])
                    # 组合后的大消息字段反序列化之后的格式和正常消息有一些差异，需要转一下格式
                    formated_data = large_msg_avro_data_transfer(new_cols)
                    self.large_record[const.NEW_COLUMNS] = formated_data
                    self.large_record[const.NEW_COLUMNS_B] = b''
                except json.JSONDecodeError as err:
                    return err

            record = self.large_record
            record[const.INDEX] = -1

        on_entry(message.partition, message.offset, valid_ps, record, trans2sql)

        # 当消费到CHECKPOINT消息时，做一次提交
        if record[const.MESSAGE_TYPE] == const.CHECKPOINT:
            offsets = {
                kafka.TopicPartition(message.topic, message.partition): kafka.OffsetAndMetadata(message.offset + 1,
                                                                                                None),
            }
            consumer.commit(offsets)
