# -*- coding:utf-8 -*-
import datetime

import const
import deserializer


def trans2sql(record, des: deserializer.RecordDeserializer):
    message_type = record[const.MESSAGE_TYPE]

    if message_type == const.HEARTBEAT:
        print("heartbeat:{}".format(record[const.HEARTBEAT_TIMESTAMP]))
    elif message_type == const.CHECKPOINT:
        print("Checkpoint: gtid={}, file={}, position={}".format(record[const.SYNCED_GTID],
                                                                 record[const.FILE_NAME],
                                                                 record[const.POSITION]))
    elif message_type == const.BEGIN:
        print("BEGIN")
    elif message_type == const.COMMIT:
        print("COMMIT")
    elif message_type == const.ROLLBACK:
        print("ROLLBACK")
    elif message_type in (const.DDL, const.INIT_DDL, const.STATEMENT):
        print(record[const.SQL])
    elif message_type == const.INSERT:
        print(trans_insert(record, des))
    elif message_type == const.INIT_INSERT:
        print(trans_init_insert(record, des))
    elif message_type == const.DELETE:
        print(trans_delete(record, des))
    elif message_type == const.UPDATE:
        print(trans_update(record, des))
    else:
        print("unsupported message type")


def trans_insert(record, des: deserializer.RecordDeserializer) -> str:
    col_values = []

    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        origType = record[const.COLUMNS][idx]["originalType"]
        d = decode(col, des.get_type(col), origType)
        col_values.append(d)

    sql = "INSERT INTO `{}`.`{}` VALUES ({})".format(
        record[const.SCHEMA_NAME],
        record[const.TABLE_NAME],
        ", ".join(col_values))

    return sql


def trans_init_insert(record, des: deserializer.RecordDeserializer) -> str:
    col_values = []
    col_names = []
    for col in record[const.COLUMNS]:
        col_names.append(col['name'])
    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        origType = record[const.COLUMNS][idx]["originalType"]
        d = decode(col, des.get_type(col), origType)
        col_values.append(d)

    sql = "INSERT INTO `{}`.`{}`({}) VALUES ({})".format(
        record[const.SCHEMA_NAME],
        record[const.TABLE_NAME],
        ", ".join(col_names),
        ", ".join(col_values))

    return sql


def trans_delete(record, des: deserializer.RecordDeserializer):
    conditions = []
    for idx, col in enumerate(record[const.OLD_COLUMNS]):
        if col is None:
            continue

        col_def = record[const.COLUMNS][idx]
        origType = record[const.COLUMNS][idx]["originalType"]
        d = decode(col, des.get_type(col), origType)

        if "NULL" == d:
            conditions.append("{} IS NULL".format(col_def['name']))
        else:
            conditions.append("{} = {}".format(col_def['name'], d))

    where = " AND ".join(conditions)
    sql = "DELETE FROM `{}`.`{}` WHERE {} LIMIT 1".format(
        record[const.SCHEMA_NAME], record[const.TABLE_NAME], where)

    return sql


def trans_update(record, des: deserializer.RecordDeserializer):
    assignments = []

    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        if col is None:
            continue

        col_def = record[const.COLUMNS][idx]
        origType = record[const.COLUMNS][idx]["originalType"]
        d = decode(col, des.get_type(col), origType)
        assignments.append("{} = {}".format(col_def['name'], d))

    conditions = []
    for idx, col in enumerate(record[const.OLD_COLUMNS]):
        if col is None:
            continue
        col_def = record[const.COLUMNS][idx]
        origType = record[const.COLUMNS][idx]["originalType"]
        d = decode(col, des.get_type(col), origType)

        if "NULL" == d:
            conditions.append("{} IS NULL".format(col_def['name']))
        else:
            conditions.append("{} = {}".format(col_def['name'], d))

    assign = ", ".join(assignments)
    where = " AND ".join(conditions)
    sql = "UPDATE `{}`.`{}` SET {} WHERE {} LIMIT 1".format(
        record[const.SCHEMA_NAME], record[const.TABLE_NAME], assign, where)

    return sql


def decode(data, dtype, origType) -> str:
    # print("decode: {}, data type: {}, original type: {}".format(data, dtype, origType))
    if data is None or "" == data:
        return 'DEFAULT'

    # 解析数值类型
    if dtype in ("Integer", "Float", "Decimal"):
        if "float" in origType.lower():
            return str(format(data['value'], '.7g'))
        return str(data['value'])

    # 解析字符串类型
    if dtype in ("Character"):
        return "x'" + data['value'].hex() + "'"

    # 解析时间戳
    if dtype in ("Timestamp"):
        secs = data['timestamp']
        millis = data['millis']
        dt = datetime.datetime.fromtimestamp(timestamp=secs + millis / 1000)
        return "_binary'" + dt.strftime('%Y-%m-%d %H:%M:%S.%f') + "'"

    # 解析时间类型
    if dtype in ("DateTime"):
        return "_binary'" + str(parse_datetime(data, None)) + "'"
    if dtype in ("TimestampWithTimeZone"):
        return "_binary'" + str(parse_datetime(data['value'], data['timezone'])) + "'"

    # 解析二进制类型
    if dtype in ("BinaryGeometry", "BinaryObject"):
        return "x'" + data['value'].hex() + "'"

    # 解析文本类型
    if dtype in ("TextGeometry", "TextObject"):
        if "json" == data['type']:
            return "'" + data['value'] + "'"
        return "_binary'" + data['value'] + "'"

    if dtype in ("EmptyObject"):
        return "NULL"

    print("unsupported data type")


# timezone 暂时不支持，带扩展
def parse_datetime(data, timezone):
    year = data['year'] if 'year' in data else None
    month = data['month'] if 'month' in data else None
    day = data['day'] if 'day' in data else None
    hour = data['hour'] if 'hour' in data else None
    minute = data['minute'] if 'minute' in data else None
    second = data['second'] if 'second' in data else None
    millis = data['millis'] if 'millis' in data else None
    if year is not None:
        if month is not None and day is not None:
            # 对应 datetime
            if hour is not None and minute is not None and second is not None:
                if timezone is not None:
                    tz = datetime.timezone(offset=datetime.timedelta(hours=timezone))
                    # 数据订阅中只保留毫秒位(millisecond)，python的datetime中是微秒位(microsecond)
                    return datetime.datetime(year, month, day, hour, minute, second, millis * 1000, tzinfo=tz) \
                        .strftime('%Y-%m-%d %H:%M:%S.%f')
                else:
                    return "%04d-%02d-%02d %02d:%02d:%02d.%03d" % (year, month, day, hour, minute, second, millis)
            else:
                # 对应 date
                return "%04d-%02d-%02d" % (year, month, day)
        else:
            # 对应 year
            return year
    else:
        # 对应 time
        if hour is not None and minute is not None and second is not None:
            return "%02d:%02d:%02d.%03d" % (hour, minute, second, millis)
        else:
            raise Exception("wrong datetime: {}".format(data))
